//  LISTBC is a simple listing program that adds a brace count
//  to the beginning of each line.  Text in comments (C++ or ANSI),
//  strings, and character constants is ignored.
//
//  LISTBC is implemented as a state machine.

#include <stdio.h>
#include <stdlib.h>
#include <dir.h>


class Buffer {

      char        line[1024];
      char *      linep;
      int         braces;
      int         parens;
      int         brackets;
      int         lineNumber;

   public:

      Buffer() {
            linep      = &line[0];
            braces     = 0;
            parens     = 0;
            brackets   = 0;
            lineNumber = 0;
            }

      void dumpLine();
      void addChar( int c ) { *linep++ = (char)c; }
      void testChar( int c );
      };


void Buffer::dumpLine()
{
      lineNumber++;

      printf( "%4d : ", lineNumber );

      if ( braces != 0 )
            printf( "{ %2d  ", braces );
      else
            fputs( "      ", stdout );

      if ( parens != 0 )
            printf( "( %2d  ", parens );
      else
            fputs( "      ", stdout );

      if ( brackets != 0 )
            printf( "[ %2d  |", brackets );
      else
            fputs( "      |", stdout );


      for ( char * lp = &line[0];  lp < linep;  lp++ )
            fputc( *lp, stdout );

      fputc( '\n', stdout );

      linep = &line[0];
}



void Buffer::testChar( int c )
{
      switch ( (char)c ) {
            case '{' : braces++; break;
            case '}' : braces--; break;
            case '[' : brackets++; break;
            case ']' : brackets--; break;
            case '(' : parens++; break;
            case ')' : parens--; break;
            }
}



void processFile( FILE * input )
{
      int         state;
      int         c;
      Buffer      buffer;


      c = 0;
      state = 0;
      while ( c != (int)EOF ) {

            c = fgetc( input );


            if ( c == '\n' )
                  buffer.dumpLine();
            else
                  buffer.addChar( c );


            switch ( state ) {

                  case 0:     //  Normal, reading chars
                        if ( c == (int)'/' )
                              state = 1;
                        else if ( c == (int)'\"' )
                              state = 5;
                        else if ( c == (int)'\'' )
                              state = 7;
                        else
                              buffer.testChar( c );

                        break;

                  case 1:     //  First '/', looking for comment
                        if ( c == (int)'/' )
                              state = 2;
                        else if ( c == (int)'*' )
                              state = 3;
                        else
                              state = 0;
                        break;

                  case 2:     //  In C++ comment, looking for EOL
                        if ( c == (int)'\n' )
                              state = 0;
                        break;

                  case 3:     //  In ANSI comment, looking for '*'
                        if ( c == (int)'*' )
                              state = 4;
                        break;

                  case 4:     //  In ANSI comment, looking for '/' to terminate comment
                        if ( c == (int)'/' )
                              state = 0;
                        else
                              state = 3;
                        break;

                  case 5:     //  In string
                        if ( c == (int)'\\' )
                              state = 6;
                        else if ( c == (int)'\"' )
                              state = 0;
                        break;
                              
                  case 6:     //  In string, skipping escaped character
                        state = 5;
                        break;

                  case 7:     //  In char constant
                        if ( c == (int)'\\' )
                              state = 8;
                        else if ( c == (int)'\'' )
                              state = 0;
                        break;

                  case 8:     //  In char constant, skipping escaped character
                        state = 7;

                  }
            }


      fclose( input );
      fputc( 12, stdout );
}


void doFile( char * name )
{
      FILE *      input;


      fputs( "###  ", stdout );
      fputs( name, stdout );
      fputs( ": ", stdout );

      input = fopen( name, "r" );
      if ( input == (FILE *)NULL ) 
            fputs( "Can't open.  ###\n", stdout );
      else {
            fputs( "Listing folows.  ###\n\n", stdout );
            processFile( input );
            }

      fclose( input );
}



int main( int argc, char ** argv )
{
      for ( int arg = 1;  arg < argc;  arg++ ) {

            struct ffblk      fblk;


            if ( findfirst( argv[arg], & fblk, 0 ) == 0 ) {

                  doFile( fblk.ff_name );
                  while ( findnext( & fblk ) == 0 )
                        doFile( fblk.ff_name );

                  }

            }

      return 0;
}

